@@ -1,7 +1,7 @@
-/* $Id: imap.c,v 1.14 2000/12/31 22:17:54 me Exp $
+/* $Id: imap.c,v 1.23 2001/02/28 01:02:50 me Exp $
  *
  * isync - IMAP4 to maildir mailbox synchronizer
- * Copyright (C) 2000 Michael R. Elkins <me@mutt.org>
+ * Copyright (C) 2000-1 Michael R. Elkins <me@mutt.org>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -120,6 +120,9 @@ verify_cert (SSL * ssl)
 static int
 init_ssl (config_t * conf)
 {
+    SSL_METHOD *method;
+    int options = 0;
+
     if (!conf->cert_file)
     {
 	puts ("Error, CertificateFile not defined");
@@ -127,7 +130,14 @@ init_ssl (config_t * conf)
     }
     SSL_library_init ();
     SSL_load_error_strings ();
-    SSLContext = SSL_CTX_new (SSLv23_client_method ());
+
+    if (conf->use_tlsv1 && !conf->use_sslv2 && !conf->use_sslv3)
+	method = TLSv1_client_method ();
+    else
+	method = SSLv23_client_method ();
+
+    SSLContext = SSL_CTX_new (method);
+
     if (access (conf->cert_file, F_OK))
     {
 	if (errno != ENOENT)
@@ -148,11 +158,13 @@ init_ssl (config_t * conf)
     }
 
     if (!conf->use_sslv2)
-	SSL_CTX_set_options (SSLContext, SSL_OP_NO_SSLv2);
+	options |= SSL_OP_NO_SSLv2;
     if (!conf->use_sslv3)
-	SSL_CTX_set_options (SSLContext, SSL_OP_NO_SSLv3);
+	options |= SSL_OP_NO_SSLv3;
     if (!conf->use_tlsv1)
-	SSL_CTX_set_options (SSLContext, SSL_OP_NO_TLSv1);
+	options |= SSL_OP_NO_TLSv1;
+
+    SSL_CTX_set_options (SSLContext, options);
 
     /* we check the result of the verification after SSL_connect() */
     SSL_CTX_set_verify (SSLContext, SSL_VERIFY_NONE, 0);
@@ -180,6 +192,43 @@ socket_write (Socket_t * sock, char *buf, size_t len)
     return write (sock->fd, buf, len);
 }
 
+static void
+socket_perror (const char *func, Socket_t *sock, int ret)
+{
+#if HAVE_LIBSSL
+    int err;
+
+    if (sock->use_ssl)
+    {
+	switch ((err = SSL_get_error (sock->ssl, ret)))
+	{
+	    case SSL_ERROR_SYSCALL:
+	    case SSL_ERROR_SSL:
+		if ((err = ERR_get_error ()) == 0)
+		{
+		    if (ret == 0)
+			fprintf (stderr, "SSL_%s:got EOF\n", func);
+		    else
+			fprintf (stderr, "SSL_%s:%d:%s\n", func,
+				errno, strerror (errno));
+		}
+		else
+		    fprintf (stderr, "SSL_%s:%d:%s\n", func, err,
+			    ERR_error_string (err, 0));
+		return;
+	    default:
+		fprintf (stderr, "SSL_%s:%d:unhandled SSL error\n", func, err);
+		break;
+	}
+	return;
+    }
+#else
+    (void) sock;
+    (void) ret;
+#endif
+    perror (func);
+}
+
 /* simple line buffering */
 static int
 buffer_gets (buffer_t * b, char **s)
@@ -215,10 +264,7 @@ buffer_gets (buffer_t * b, char **s)
 
 	    if (n <= 0)
 	    {
-		if (n == -1)
-		    perror ("read");
-		else
-		    puts ("EOF");
+		socket_perror ("read", b->sock, n);
 		return -1;
 	    }
 
@@ -367,6 +413,7 @@ imap_exec (imap_t * imap, const char *fmt, ...)
     char *cmd;
     char *arg;
     char *arg1;
+    int n;
 
     va_start (ap, fmt);
     vsnprintf (tmp, sizeof (tmp), fmt, ap);
@@ -375,7 +422,12 @@ imap_exec (imap_t * imap, const char *fmt, ...)
     snprintf (buf, sizeof (buf), "%d %s\r\n", ++Tag, tmp);
     if (Verbose)
 	fputs (buf, stdout);
-    socket_write (imap->sock, buf, strlen (buf));
+    n = socket_write (imap->sock, buf, strlen (buf));
+    if (n <= 0)
+    {
+	socket_perror ("write", imap->sock, n);
+	return -1;
+    }
 
     for (;;)
     {
@@ -459,10 +511,20 @@ imap_exec (imap_t * imap, const char *fmt, ...)
 	    }
 	    resp = cram (cmd, imap->box->user, imap->box->pass);
 
-	    socket_write (imap->sock, resp, strlen (resp));
+	    n = socket_write (imap->sock, resp, strlen (resp));
+	    if (n <= 0)
+	    {
+		socket_perror ("write", imap->sock, n);
+		return -1;
+	    }
 	    if (Verbose)
 		puts (resp);
-	    socket_write (imap->sock, "\r\n", 2);
+	    n = socket_write (imap->sock, "\r\n", 2);
+	    if (n <= 0)
+	    {
+		socket_perror ("write", imap->sock, n);
+		return -1;
+	    }
 	    free (resp);
 	    imap->cram = 0;
 	}
@@ -496,7 +558,6 @@ imap_open (config_t * box, unsigned int minuid, imap_t * imap)
     int s;
     struct sockaddr_in sin;
     struct hostent *he;
-    char *ns_prefix = "";
     int reuse = 0;
 #if HAVE_LIBSSL
     int use_ssl = 0;
@@ -542,6 +603,7 @@ imap_open (config_t * box, unsigned int minuid, imap_t * imap)
 
     imap->box = box;
     imap->minuid = minuid;
+    imap->prefix = "";
 
     if (!reuse)
     {
@@ -586,7 +648,9 @@ imap_open (config_t * box, unsigned int minuid, imap_t * imap)
 	if (!reuse)
 	{
 #if HAVE_LIBSSL
-	    if (!box->use_imaps)
+	    if (box->use_imaps)
+		use_ssl = 1;
+	    else if (box->use_sslv2 || box->use_sslv3 || box->use_tlsv1)
 	    {
 		/* let's see what this puppy can do... */
 		if ((ret = imap_exec (imap, "CAPABILITY")))
@@ -600,22 +664,20 @@ imap_open (config_t * box, unsigned int minuid, imap_t * imap)
 		    use_ssl = 1;
 		}
 
-		if (!use_ssl)
+	    }
+
+	    if (!use_ssl)
+	    {
+		if (box->require_ssl)
 		{
-		    if (box->require_ssl)
-		    {
-			puts ("Error, SSL support not available");
-			ret = -1;
-			break;
-		    }
-		    else
-			puts ("Warning, SSL support not available");
+		    puts ("Error, SSL support not available");
+		    ret = -1;
+		    break;
 		}
+		else
+		    puts ("Warning, SSL support not available");
 	    }
 	    else
-		use_ssl = 1;
-
-	    if (use_ssl)
 	    {
 		/* initialize SSL */
 		if (init_ssl (box))
@@ -629,10 +691,7 @@ imap_open (config_t * box, unsigned int minuid, imap_t * imap)
 		ret = SSL_connect (imap->sock->ssl);
 		if (ret <= 0)
 		{
-		    ret = SSL_get_error (imap->sock->ssl, ret);
-		    printf ("Error, SSL_connect: %s\n",
-			    ERR_error_string (ret, 0));
-		    ret = -1;
+		    socket_perror ("connect", imap->sock, ret);
 		    break;
 		}
 
@@ -640,12 +699,18 @@ imap_open (config_t * box, unsigned int minuid, imap_t * imap)
 		if ((ret = verify_cert (imap->sock->ssl)))
 		    break;
 
+		/* to conform to RFC2595 we need to forget all information
+		 * retrieved from CAPABILITY invocations before STARTTLS.
+		 */
+		imap->have_namespace = 0;
+		imap->have_cram = 0;
+		imap->have_starttls = 0;
+
 		imap->sock->use_ssl = 1;
 		puts ("SSL support enabled");
 
-		if (box->use_imaps)
-		    if ((ret = imap_exec (imap, "CAPABILITY")))
-			break;
+		if ((ret = imap_exec (imap, "CAPABILITY")))
+		    break;
 	    }
 #else
 	    if ((ret = imap_exec (imap, "CAPABILITY")))
@@ -680,7 +745,10 @@ imap_open (config_t * box, unsigned int minuid, imap_t * imap)
 		    (ret =
 		     imap_exec (imap, "LOGIN \"%s\" \"%s\"", box->user,
 				box->pass)))
+		{
+		    puts ("Error, LOGIN failed");
 		    break;
+		}
 	    }
 
 	    /* get NAMESPACE info */
@@ -696,12 +764,14 @@ imap_open (config_t * box, unsigned int minuid, imap_t * imap)
 	    is_list (imap->ns_personal->child) &&
 	    is_atom (imap->ns_personal->child->child))
 	{
-	    ns_prefix = imap->ns_personal->child->child->val;
+	    imap->prefix = imap->ns_personal->child->child->val;
 	}
 
 	fputs ("Selecting mailbox... ", stdout);
 	fflush (stdout);
-	if ((ret = imap_exec (imap, "SELECT %s%s", ns_prefix, box->box)))
+	if (
+	    (ret =
+	     imap_exec (imap, "SELECT \"%s%s\"", imap->prefix, box->box)))
 	    break;
 	printf ("%d messages, %d recent\n", imap->count, imap->recent);
 
@@ -727,7 +797,6 @@ imap_open (config_t * box, unsigned int minuid, imap_t * imap)
 void
 imap_close (imap_t * imap)
 {
-    puts ("Closing IMAP connection");
     imap_exec (imap, "LOGOUT");
     close (imap->sock->fd);
     free (imap->sock);
@@ -755,22 +824,30 @@ write_strip (int fd, char *buf, size_t len)
     return 0;
 }
 
-static void
+static int
 send_server (Socket_t * sock, const char *fmt, ...)
 {
     char buf[128];
     char cmd[128];
     va_list ap;
+    int n;
 
     va_start (ap, fmt);
     vsnprintf (buf, sizeof (buf), fmt, ap);
     va_end (ap);
 
     snprintf (cmd, sizeof (cmd), "%d %s\r\n", ++Tag, buf);
-    socket_write (sock, cmd, strlen (cmd));
+    n = socket_write (sock, cmd, strlen (cmd));
+    if (n <= 0)
+    {
+	socket_perror ("write", sock, n);
+	return -1;
+    }
 
     if (Verbose)
 	fputs (cmd, stdout);
+
+    return 0;
 }
 
 int
@@ -848,10 +925,7 @@ imap_fetch_message (imap_t * imap, unsigned int uid, int fd)
 		}
 		else
 		{
-		    if (n == (size_t) - 1)
-			perror ("read");
-		    else
-			puts ("EOF");
+		    socket_perror ("read", imap->sock, n);
 		    return -1;
 		}
 	    }
@@ -903,3 +977,173 @@ imap_expunge (imap_t * imap)
 {
     return imap_exec (imap, "EXPUNGE");
 }
+
+int
+imap_copy_message (imap_t * imap, unsigned int uid, const char *mailbox)
+{
+    return imap_exec (imap, "UID COPY %u \"%s%s\"", uid, imap->prefix,
+		      mailbox);
+}
+
+int
+imap_append_message (imap_t * imap, int fd, message_t * msg)
+{
+    char buf[1024];
+    size_t len;
+    size_t sofar = 0;
+    int lines = 0;
+    char flagstr[128];
+    char *s;
+    size_t i;
+    size_t start, end;
+    char *arg;
+
+    /* ugh, we need to count the number of newlines */
+    while (sofar < msg->size)
+    {
+	len = msg->size - sofar;
+	if (len > sizeof (buf))
+	    len = sizeof (buf);
+	len = read (fd, buf, len);
+	if (len == (size_t) - 1)
+	{
+	    perror ("read");
+	    return -1;
+	}
+	for (i = 0; i < len; i++)
+	    if (buf[i] == '\n')
+		lines++;
+	sofar += len;
+    }
+
+    flagstr[0] = 0;
+    if (msg->flags)
+    {
+	strcpy (flagstr, "(");
+	if (msg->flags & D_DELETED)
+	    snprintf (flagstr + strlen (flagstr),
+		      sizeof (flagstr) - strlen (flagstr), "%s\\Deleted",
+		      flagstr[1] ? " " : "");
+	if (msg->flags & D_ANSWERED)
+	    snprintf (flagstr + strlen (flagstr),
+		      sizeof (flagstr) - strlen (flagstr), "%s\\Answered",
+		      flagstr[1] ? " " : "");
+	if (msg->flags & D_SEEN)
+	    snprintf (flagstr + strlen (flagstr),
+		      sizeof (flagstr) - strlen (flagstr), "%s\\Seen",
+		      flagstr[1] ? " " : "");
+	if (msg->flags & D_FLAGGED)
+	    snprintf (flagstr + strlen (flagstr),
+		      sizeof (flagstr) - strlen (flagstr), "%s\\Flagged",
+		      flagstr[1] ? " " : "");
+	if (msg->flags & D_DRAFT)
+	    snprintf (flagstr + strlen (flagstr),
+		      sizeof (flagstr) - strlen (flagstr), "%s\\Draft",
+		      flagstr[1] ? " " : "");
+	snprintf (flagstr + strlen (flagstr),
+		  sizeof (flagstr) - strlen (flagstr), ") ");
+    }
+
+    send_server (imap->sock, "APPEND %s%s %s{%d}",
+		 imap->prefix, imap->box->box, flagstr, msg->size + lines);
+
+    if (buffer_gets (imap->buf, &s))
+	return -1;
+    if (Verbose)
+	puts (s);
+
+    if (*s != '+')
+    {
+	puts ("Error, expected `+' from server (aborting)");
+	return -1;
+    }
+
+    /* rewind */
+    lseek (fd, 0, 0);
+
+    sofar = 0;
+    while (sofar < msg->size)
+    {
+	len = msg->size - sofar;
+	if (len > sizeof (buf))
+	    len = sizeof (buf);
+	len = read (fd, buf, len);
+	if (len == (size_t) - 1)
+	    return -1;
+	start = 0;
+	while (start < len)
+	{
+	    end = start;
+	    while (end < len && buf[end] != '\n')
+		end++;
+	    if (start != end)
+		socket_write (imap->sock, buf + start, end - start);
+	    /* only send a crlf if we actually hit the end of a line.  we
+	     * might be in the middle of a line in which case we don't
+	     * send one.
+	     */
+	    if (end != len)
+		socket_write (imap->sock, "\r\n", 2);
+	    start = end + 1;
+	}
+	sofar += len;
+    }
+    socket_write (imap->sock, "\r\n", 2);
+
+    for (;;)
+    {
+	if (buffer_gets (imap->buf, &s))
+	    return -1;
+
+	if (Verbose)
+	    puts (s);
+
+	arg = next_arg (&s);
+	if (*arg == '*')
+	{
+	    /* XXX just ignore it for now */
+	}
+	else if (atoi (arg) != (int) Tag)
+	{
+	    puts ("wrong tag");
+	    return -1;
+	}
+	else
+	{
+	    int uid;
+
+	    arg = next_arg (&s);
+	    if (strcmp (arg, "OK"))
+		return -1;
+	    arg = next_arg (&s);
+	    if (*arg != '[')
+		break;
+	    arg++;
+	    if (strcasecmp ("APPENDUID", arg))
+	    {
+		puts ("Error, expected APPENDUID");
+		break;
+	    }
+	    arg = next_arg (&s);
+	    if (!arg)
+		break;
+	    if (atoi (arg) != (int) imap->uidvalidity)
+	    {
+		puts ("Error, UIDVALIDITY doesn't match APPENDUID");
+		return -1;
+	    }
+	    arg = next_arg (&s);
+	    if (!arg)
+		break;
+	    uid = strtol (arg, &s, 10);
+	    if (*s != ']')
+	    {
+		/* parse error */
+		break;
+	    }
+	    return uid;
+	}
+    }
+
+    return 0;
+}
